/*******************************************************************************
*                                                                              *
*   (C) 1997-2012 by Ernst W. Mayer.                                           *
*                                                                              *
*  This program is free software; you can redistribute it and/or modify it     *
*  under the terms of the GNU General Public License as published by the       *
*  Free Software Foundation; either version 2 of the License, or (at your      *
*  option) any later version.                                                  *
*                                                                              *
*  This program is distributed in the hope that it will be useful, but WITHOUT *
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
*  more details.                                                               *
*                                                                              *
*  You should have received a copy of the GNU General Public License along     *
*  with this program; see the file GPL.txt.  If not, you may view one at       *
*  http://www.fsf.org/licenses/licenses.html, or obtain one by writing to the  *
*  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA     *
*  02111-1307, USA.                                                            *
*                                                                              *
*******************************************************************************/

#include <errno.h> 

// package via tar -cvzf factor_qmmp.tgz factor_qmmp*.c masterdefs.h platform.h types.h int_macros.h
#include "masterdefs.h"
#include "types.h"	// This includes platform.h
#include "int_macros.h"

#define NUM_SIEVING_PRIME 100000	// 1e5 is ~optimal on x86_64 for sieving to ~1T

/* Build & usage: First do sieve-only factor.c run for the MMp in question, printing out the k's which pass the small-p sieve.

g42 -o factor_qmmp -Wall -O3 -lm factor_qmmp*.c

Test #1: This should find the small-prime factor 582994261 in around 20 GHz-seconds on x86_64:

	time ./factor_qmmp -p 43112609 -k 5 -imax 10000000000

Test #2: This should find the small-prime factor 7450239943 in around 20 GHz-seconds on x86_64:

	time ./factor_qmmp -p 43112609 -imin 7000000000 -imax 8000000000 -k 665

*/

#undef YES_ASM
#if(defined(CPU_TYPE_AMD64) && defined(COMPILER_TYPE_GCC) && (OS_BITS == 64))
	#define YES_ASM
#endif

/******************/

void ASSERT(long line, char*file, int expr, char*assert_string)
{
	/* Define a convenient spot to set a breakpoint: */
	if(!expr)
	{
		fprintf(stderr,"ERROR: at line %lu of file %s\n", line, file);
		fprintf(stderr,"Assertion failed: %s\n", assert_string);
		/* Flush all output streams prior to asserting.
		We replace the original assert(0) call with an exit(EXIT_FAILURE),
		since some compilers seem to like to optimize away assertions. */
		fflush(NULL);
		exit(EXIT_FAILURE);
	}
}

// This tries to be clever and uses a divide-and-conquer scheme:
uint32 leadz32(uint32 i)
{
	uint32 lz, k, shift, ones_mask = 0xFFFFFFFF;
	if(i == 0) return 32;
	if((int32)i < 0) return 0;

	lz    =  0;
	shift = 16;
	k     = 16;
	while(k > 0)
	{
		if( (ones_mask >> shift) < i )
		{
			k >>= 1;
			shift -= k;
		}
		else
		{
			lz += k;
			k >>= 1;
			shift += k;
		}
	}
	return lz;
}

// This uses linear bit-checking:
uint32 trailz32(uint32 x)
{
	uint32 i;
	if(x == 0) return 32;
	for(i = 1; i < 32; i++)
	{
		if(x & 1)break;
		x >>= 1;
	}
	return(i-1);
}

uint64 ibits64(uint64 i, uint32 beg, uint32 nbits)
{
	uint64 ib;
	uint64 ones_mask = 0xFFFFFFFFFFFFFFFFull;
	ib = (i >> beg) & ~(ones_mask << nbits);
	return ( ib );
}

uint32 leadz64(uint64 i)
{
	uint32 lz;
	if(i == 0) return 64;
	if(( int64)i < 0) return 0;
#ifdef X64_ASM
	int bpos;
	__asm__ volatile (\
		"bsrq %[__i],%%rax		\n\t"\
		"movl %%eax,%[__bpos]	\n\t"\
		:	/* outputs: none */\
		: [__i] "m" (i)	/* All inputs from memory addresses here */\
		 ,[__bpos] "m" (bpos)	\
		: "cc","memory","rax"	/* Clobbered registers */\
		);
	lz = (63 - bpos);	// BSR returns *index* of leftmost set bit, must subtract from (#bits - 1) to get #lz.
#else
	uint32 k, shift;
	uint64 ones_mask = 0xFFFFFFFFFFFFFFFFull;
	lz    =  0;
	shift = 32;
	k     = 32;
	while(k > 0)
	{
		if( (ones_mask >> shift) < i )
		{
			k >>= 1;
			shift -= k;
		}
		else
		{
			lz += k;
			k >>= 1;
			shift += k;
		}
	}
	DBG_ASSERT(HERE, ( (i << lz) >> lz == i ),"ERROR A in leadz64");
	DBG_ASSERT(HERE, ( i >> (64-lz)    == 0 ),"ERROR B in leadz64");
#endif
	return lz;
}

uint32 trailz64(uint64 x)
{
	if(x == 0) return 64;
#ifdef X64_ASM
	int bpos;
	__asm__ volatile (\
		"bsfq %[__x],%%rax		\n\t"\
		"movl %%eax,%[__bpos]	\n\t"\
		:	/* outputs: none */\
		: [__x] "m" (x)	/* All inputs from memory addresses here */\
		 ,[__bpos] "m" (bpos)	\
		: "cc","memory","rax"	/* Clobbered registers */\
		);
	return bpos;
#else
	uint32 i;
	const uint64 one64 = 1;
	for(i = 1; i < 64; i++)
	{
		if(x & one64) {
			break;
		}
		x >>= 1;
	}
	return(i-1);
#endif
}

uint64 gcd64(uint64 x, uint64 y)
{
	uint64 q, f;
	if(y == 0) {
		return x;
	}
	while(y) {
		q = x/y;
		f = x - q*y;
		x = y;
		y = f;
	}
	return x;
}

int twopmodq32(uint32 p, uint32 q)	// (2^-p % q) == 0
{
	 int32 j;
	uint32 lead5, pshift, qhalf, qinv, zshift, start_index, x, lo, hi;

	ASSERT(HERE, (q&1) == 1, "twopmodq32: even modulus!");
	qhalf = q >> 1;	/* = (q-1)/2, since q odd. */

	pshift = p + 32;
	if(pshift < p)	/* Need special-casing for p just below 2^32  - the primes 2^32-(5,17) are good testcases here. */
	{
		j = -1;	/* leadz32(pshift) for 33-bit pshift goes negative */
		/* Extract leftmost 5 bits of pshift: */
		lead5 = 16 + (pshift >> 28);
	}
	else
	{
		/* Find number of leading zeros in p, use it to find the position of the leftmost ones bit: */
		j = leadz32(pshift);
		/* Extract leftmost 5 bits of pshift: */
		lead5 = ((pshift<<j) >> 27);
	}

	start_index = 32-j-5;	/* Leftward bit at which to start the l-r binary powering, assuming
							the leftmost 5 bits have already been processed via a shift (see next). */

	zshift = 31 - lead5;
	zshift <<= 1;		/* Doubling the shift count here takes cares of the first SQR_LOHI */
	pshift = ~pshift;	/* Overflow doesn't matter here, as long as we got the leading 5 bits of pshift right. */

	/*
	!    Find modular inverse (mod 2^32) of q in preparation for modular multiply.
	!    We use the simple and elegant iterative inversion method of Montgomery,
	!    which amounts to a modular analogue of Newton's method for iterative inversion:
	!
	!    0)   Zinv = Z                   ! Z*Zinv == 1 (mod 2^3)
	!    1)   Zinv = Zinv*(2 - Z*Zinv)   ! Z*Zinv == 1 (mod 2^6), etc.
	!
	!    where the number of correct bits (at the low end) doubles at each step,
	!    all arithmetic is modulo 2^32 and we repeat step (1) until we have the needed 32 bits.
	!
	!    We choose a different starting value of Zinv, XOR(3*Z, 2),
	!    so the first congruence holds modulo 2^4, thus requiring just 3 iterations.
	*/
	qinv = (q+q+q) ^ (uint32)2;	/* Overflow doesn't matter here, since we only care about the low 2 bits of 3*q. */

	qinv = qinv*((uint32)2 - q*qinv);
	qinv = qinv*((uint32)2 - q*qinv);
	qinv = qinv*((uint32)2 - q*qinv);

	/* Since zstart is a power of two < 2^32, use a streamlined code sequence for the first iteration: */
	j = start_index-1;

	/* For 64-bit hardware, Make sure we get a 32-bit shift result here by ANDing with 2^32-1: */
	lo = (qinv << zshift) & (uint32)0xffffffff;
	/* Emulate MULH64 here by getting full 64-bit product and right-shifting: */
	lo = (uint32)(((uint64)q * (uint64)lo) >> 32);
	x  = q - lo;

	if((pshift >> j) & (uint32)1)
	{
		DBG_ASSERT(HERE, x < q,"util.c: x < q");
		/* Combines overflow-on-add and need-to-subtract-q-from-sum checks */
		if(x > qhalf) 
		{
			x += x;
			x -= q;
		}
		else
		{
			x += x;
		}
	}

	for(j = start_index-2; j >= 0; j--)
	{
		/* SQR_LOHI64(x,lo,hi): */
		MUL_LOHI32(x,x, lo,hi);
		lo *= qinv;
		MULH32(q,lo, lo);

		/* Branchless version is much faster: */
		x = hi - lo + ((-(hi < lo)) & q);

		if((pshift >> j) & (uint32)1)
		{
			/* Branchless version is much faster: */
			x = x + x - ((-(x > qhalf)) & q);
		}
	}

	/*...Double and return.	These are specialized for the case where 2^p == 1 mod q implies divisibility, in which case x = (q+1)/2. */
	return((int)((x + x - q) == 1));
}

/* For an exponent p and a factor index k, both mod 60, returns the factoring
pass number (in unit-offset form, i.e. 1-16) on which the factor should occur
if it's one of the valid combinations of p%60 and k%60. If invalid, returns 0. */
uint32 CHECK_PKMOD60(uint32 p, uint32 k)
{
ASSERT(HERE, p < 60 && k < 60, "CHECK_PKMOD60: args must both be (mod 60)!");
if(p== 1){ if(k==0)return 16; else if(k== 3)return 1; else if(k== 8)return 2; else if(k==11)return 3; else if(k==15)return 4; else if(k==20)return 5; else if(k==23)return 6; else if(k==24)return 7; else if(k==35)return 8; else if(k==36)return 9; else if(k==39)return 10; else if(k==44)return 11; else if(k==48)return 12; else if(k==51)return 13; else if(k==56)return 14; else if(k==59)return 15; else return 0; };
if(p== 7){ if(k==0)return 16; else if(k== 5)return 1; else if(k== 8)return 2; else if(k== 9)return 3; else if(k==12)return 4; else if(k==17)return 5; else if(k==20)return 6; else if(k==24)return 7; else if(k==29)return 8; else if(k==32)return 9; else if(k==33)return 10; else if(k==44)return 11; else if(k==45)return 12; else if(k==48)return 13; else if(k==53)return 14; else if(k==57)return 15; else return 0; };
if(p==11){ if(k==0)return 16; else if(k== 1)return 1; else if(k== 4)return 2; else if(k== 9)return 3; else if(k==13)return 4; else if(k==16)return 5; else if(k==21)return 6; else if(k==24)return 7; else if(k==25)return 8; else if(k==28)return 9; else if(k==33)return 10; else if(k==36)return 11; else if(k==40)return 12; else if(k==45)return 13; else if(k==48)return 14; else if(k==49)return 15; else return 0; };
if(p==13){ if(k==0)return 16; else if(k== 3)return 1; else if(k== 8)return 2; else if(k==11)return 3; else if(k==12)return 4; else if(k==15)return 5; else if(k==20)return 6; else if(k==23)return 7; else if(k==27)return 8; else if(k==32)return 9; else if(k==35)return 10; else if(k==36)return 11; else if(k==47)return 12; else if(k==48)return 13; else if(k==51)return 14; else if(k==56)return 15; else return 0; };
if(p==17){ if(k==0)return 16; else if(k== 3)return 1; else if(k== 4)return 2; else if(k== 7)return 3; else if(k==12)return 4; else if(k==15)return 5; else if(k==19)return 6; else if(k==24)return 7; else if(k==27)return 8; else if(k==28)return 9; else if(k==39)return 10; else if(k==40)return 11; else if(k==43)return 12; else if(k==48)return 13; else if(k==52)return 14; else if(k==55)return 15; else return 0; };
if(p==19){ if(k==0)return 16; else if(k== 5)return 1; else if(k== 9)return 2; else if(k==12)return 3; else if(k==17)return 4; else if(k==20)return 5; else if(k==21)return 6; else if(k==24)return 7; else if(k==29)return 8; else if(k==32)return 9; else if(k==36)return 10; else if(k==41)return 11; else if(k==44)return 12; else if(k==45)return 13; else if(k==56)return 14; else if(k==57)return 15; else return 0; };
if(p==23){ if(k==0)return 16; else if(k== 1)return 1; else if(k==12)return 2; else if(k==13)return 3; else if(k==16)return 4; else if(k==21)return 5; else if(k==25)return 6; else if(k==28)return 7; else if(k==33)return 8; else if(k==36)return 9; else if(k==37)return 10; else if(k==40)return 11; else if(k==45)return 12; else if(k==48)return 13; else if(k==52)return 14; else if(k==57)return 15; else return 0; };
if(p==29){ if(k==0)return 16; else if(k== 4)return 1; else if(k== 7)return 2; else if(k==12)return 3; else if(k==15)return 4; else if(k==16)return 5; else if(k==19)return 6; else if(k==24)return 7; else if(k==27)return 8; else if(k==31)return 9; else if(k==36)return 10; else if(k==39)return 11; else if(k==40)return 12; else if(k==51)return 13; else if(k==52)return 14; else if(k==55)return 15; else return 0; };
if(p==31){ if(k==0)return 16; else if(k== 5)return 1; else if(k== 8)return 2; else if(k== 9)return 3; else if(k==20)return 4; else if(k==21)return 5; else if(k==24)return 6; else if(k==29)return 7; else if(k==33)return 8; else if(k==36)return 9; else if(k==41)return 10; else if(k==44)return 11; else if(k==45)return 12; else if(k==48)return 13; else if(k==53)return 14; else if(k==56)return 15; else return 0; };
if(p==37){ if(k==0)return 16; else if(k== 3)return 1; else if(k== 8)return 2; else if(k==12)return 3; else if(k==15)return 4; else if(k==20)return 5; else if(k==23)return 6; else if(k==24)return 7; else if(k==27)return 8; else if(k==32)return 9; else if(k==35)return 10; else if(k==39)return 11; else if(k==44)return 12; else if(k==47)return 13; else if(k==48)return 14; else if(k==59)return 15; else return 0; };
if(p==41){ if(k==0)return 16; else if(k== 3)return 1; else if(k== 4)return 2; else if(k==15)return 3; else if(k==16)return 4; else if(k==19)return 5; else if(k==24)return 6; else if(k==28)return 7; else if(k==31)return 8; else if(k==36)return 9; else if(k==39)return 10; else if(k==40)return 11; else if(k==43)return 12; else if(k==48)return 13; else if(k==51)return 14; else if(k==55)return 15; else return 0; };
if(p==43){ if(k==0)return 16; else if(k== 5)return 1; else if(k== 8)return 2; else if(k==12)return 3; else if(k==17)return 4; else if(k==20)return 5; else if(k==21)return 6; else if(k==32)return 7; else if(k==33)return 8; else if(k==36)return 9; else if(k==41)return 10; else if(k==45)return 11; else if(k==48)return 12; else if(k==53)return 13; else if(k==56)return 14; else if(k==57)return 15; else return 0; };
if(p==47){ if(k==0)return 16; else if(k== 4)return 1; else if(k== 9)return 2; else if(k==12)return 3; else if(k==13)return 4; else if(k==24)return 5; else if(k==25)return 6; else if(k==28)return 7; else if(k==33)return 8; else if(k==37)return 9; else if(k==40)return 10; else if(k==45)return 11; else if(k==48)return 12; else if(k==49)return 13; else if(k==52)return 14; else if(k==57)return 15; else return 0; };
if(p==49){ if(k==0)return 16; else if(k==11)return 1; else if(k==12)return 2; else if(k==15)return 3; else if(k==20)return 4; else if(k==24)return 5; else if(k==27)return 6; else if(k==32)return 7; else if(k==35)return 8; else if(k==36)return 9; else if(k==39)return 10; else if(k==44)return 11; else if(k==47)return 12; else if(k==51)return 13; else if(k==56)return 14; else if(k==59)return 15; else return 0; };
if(p==53){ if(k==0)return 16; else if(k== 3)return 1; else if(k== 7)return 2; else if(k==12)return 3; else if(k==15)return 4; else if(k==16)return 5; else if(k==27)return 6; else if(k==28)return 7; else if(k==31)return 8; else if(k==36)return 9; else if(k==40)return 10; else if(k==43)return 11; else if(k==48)return 12; else if(k==51)return 13; else if(k==52)return 14; else if(k==55)return 15; else return 0; };
if(p==59){ if(k==0)return 16; else if(k== 1)return 1; else if(k== 4)return 2; else if(k== 9)return 3; else if(k==12)return 4; else if(k==16)return 5; else if(k==21)return 6; else if(k==24)return 7; else if(k==25)return 8; else if(k==36)return 9; else if(k==37)return 10; else if(k==40)return 11; else if(k==45)return 12; else if(k==49)return 13; else if(k==52)return 14; else if(k==57)return 15; else return 0; };
return 0;
}

// Montgomery-mul macros:
#define MONT_SQR64(__x,__q,__qinv,__z)\
{\
	uint64 lo,hi;					\
	SQR_LOHI64(__x,lo,hi);			\
	MULL64(__qinv,lo,lo);			\
	MULH64(__q,lo,lo);				\
	__z = hi - lo + ((-(int64)(hi < lo)) & __q);	/* did we have a borrow from (hi-lo)? */\
}

#define MONT_MUL64(__x,__y,__q,__qinv,__z)\
{\
	uint64 lo,hi;					\
	MUL_LOHI64(__x,__y,lo,hi);		\
	MULL64(__qinv,lo,lo);			\
	MULH64(__q,lo,lo);			\
	__z = hi - lo + ((-(int64)(hi < lo)) & __q);	/* did we have a borrow from (hi-lo)? */\
}

// "Downshift MUL" by unity to obtain x * B^(-1) mod q:
#define MONT_UNITY_MUL64(__x,__q,__qinv,__z)\
{\
	uint64 lo = __x;				\
	MULL64(__qinv,lo,lo);			\
	MULH64(__q,lo,lo);				\
	__z = - lo + ((-(int64)(lo != 0)) & __q);	/* did we have a borrow from (0-lo)? */\
}

#ifndef YES_ASM	/* Use x86_64-optimized asm version if available */

	#define MONT_SQR64_q4(__x0,__x1,__x2,__x3,__q0,__q1,__q2,__q3,__qinv0,__qinv1,__qinv2,__qinv3,__z0,__z1,__z2,__z3)\
	{\
		uint64 lo0,lo1,lo2,lo3,hi0,hi1,hi2,hi3;					\
		SQR_LOHI64(__x0, lo0, hi0);								\
		SQR_LOHI64(__x1, lo1, hi1);								\
		SQR_LOHI64(__x2, lo2, hi2);								\
		SQR_LOHI64(__x3, lo3, hi3);								\
		MULL64(__qinv0,lo0,lo0);								\
		MULL64(__qinv1,lo1,lo1);								\
		MULL64(__qinv2,lo2,lo2);								\
		MULL64(__qinv3,lo3,lo3);								\
		MULH64(__q0,lo0,lo0);									\
		MULH64(__q1,lo1,lo1);									\
		MULH64(__q2,lo2,lo2);									\
		MULH64(__q3,lo3,lo3);									\
		/* did we have a borrow from (hi-lo)? */				\
		__z0 = hi0 - lo0 + ((-(int64)(hi0 < lo0)) & __q0);		\
		__z1 = hi1 - lo1 + ((-(int64)(hi1 < lo1)) & __q1);		\
		__z2 = hi2 - lo2 + ((-(int64)(hi2 < lo2)) & __q2);		\
		__z3 = hi3 - lo3 + ((-(int64)(hi3 < lo3)) & __q3);		\
	}

#endif

uint64 radix_power64(const uint64 q, const uint64 qinv, uint32 n)
{
	int i,j,bmap,p;	// i and j must be signed
	uint64 iquo,lo,hi,rem64,itmp64;
	double fquo,fqinv;
	const double TWO48FLOAT = (double)1.0*0x01000000*0x01000000;
	const double TWO64FLOAT = (double)4.0*0x80000000*0x80000000;

	DBG_ASSERT(HERE, (q & 1) == 0, "modulus must be odd!");
	if(n == 0) return 1ull;
	if(n == 1) return 0ull;

	/*** The minimum nontrivial power of B we need is B^2, so first obtain that. ***/

	// Floating-point (approximate) inverse:
	fqinv = 1.0/q;
	if(q < (1ull << 48)) {	// q < 2^48
	// Method A: Use mod-doublings to get 2^72 (mod q), followed by 3 MONT_SQR64:
	#if 0
		itmp64 = 0x8000000000000000ull % q;	// 2^63 % q
		// 5 mod-doublings yield 2^68 (mod q)
		itmp64 = itmp64 + itmp64 - q;	itmp64 += (-((int64)itmp64 < 0) & q);
		itmp64 = itmp64 + itmp64 - q;	itmp64 += (-((int64)itmp64 < 0) & q);
		itmp64 = itmp64 + itmp64 - q;	itmp64 += (-((int64)itmp64 < 0) & q);
		itmp64 = itmp64 + itmp64 - q;	itmp64 += (-((int64)itmp64 < 0) & q);
		itmp64 = itmp64 + itmp64 - q;	itmp64 += (-((int64)itmp64 < 0) & q);
		MONT_SQR64(itmp64,q,qinv,itmp64);	// 2^(2*68-64) == 2^72 (mod q)
		MONT_SQR64(itmp64,q,qinv,itmp64);	// 2^80 (mod q)
		MONT_SQR64(itmp64,q,qinv,itmp64);	// 2^96 (mod q)
	#else
	// Method B uses fast floating-point mod:
		fquo = TWO48FLOAT*fqinv;
		iquo = (uint64)fquo;
		itmp64 = TWO48FLOAT - q*iquo;	// 2^48 % q, which may be as large as 48 bits
		SQR_LOHI64(itmp64, lo,hi);
		fquo = (hi*TWO64FLOAT + lo)*fqinv;
		iquo = (uint64)fquo;	// Need both of these for check below
		itmp64 = lo - q*iquo;	// Only need low 64 bits of difference
		// Floating-point method here fails in a tiny fraction of cases - see notes in "else" block for details:
		if(itmp64 > q)
		{
			// This check allows us to differentiate between incorrect upward-rounded and (rarer) downward-rounded cases:
			if(DNINT(fquo) == (double)iquo) {	// Incorrect   upward-rounded
				itmp64 += q;	//	printf("A...");
			} else {							// Incorrect downward-rounded
				itmp64 -= q;	//	printf("B...");
			}
		}
	#endif
	}
	else	// q in [2^48, 2^64)
	{
		fquo = TWO48FLOAT*TWO48FLOAT*fqinv;
		iquo = (uint64)fquo;
		// Bottom 64 bits of 2^96 - q*(2^96/q)
		itmp64 = -(iquo * q);
		// Floating-point method here fails in a tiny fraction of cases (which on x86 is highly dependent on build mode, likely due to in-register
		// precision effects: in my test, ~35000 of 10^9 fail for debug-build, only 19 for opt-build), but these failures easily spotted via this check.
		// If quotient has (exact) fractional part = 0.99...., may end up with fquo = ceiling(2^96/q) rather than floor; thus want q - (2^64 - itmp64):
		if(itmp64 > q)
		{
			// This check allows us to differentiate between incorrect upward-rounded and (rarer) downward-rounded cases:
			if(DNINT(fquo) == (double)iquo) {	// Incorrect   upward-rounded, e.g. fquo = 1084809392143.0001, exact = 1084809392142.999...
			//	printf("radix_power64: q = %20llu < itmp64 = (int64)%20lld, fquo = %20.4f, (double)rem64 = %20.4f\n",q,(int64)itmp64, fquo, (double)rem64);
				itmp64 += q;	//	printf("C...");
			} else {							// Incorrect downward-rounded, e.g. fquo = 7344640876302.9990, exact = 7344640876303.0000002...
			//	printf("radix_power64: q = %20llu < itmp64 = (int64)%20lld, fquo = %20.4f *** Bad Downward ***\n",q,(int64)itmp64, fquo);
				itmp64 -= q;	//	printf("D...");
			}
		}
	}

	// Now that have B^(3/2), do a Mont-square to get B^2 % q:
	MONT_SQR64(itmp64,q,qinv,rem64);

	/* rem64 holds B^2 mod q - Now compute sequence of powers needed to obtain B^len mod q via Montgomery-muls: */
	p = n;	// Add 1 since used high-MUL version of the scaled-remainder algo ( = Algorithm A in the paper)
	// If p == 2, rem64 already contains the needed power
	if(p == 3) {
		MONT_SQR64(rem64,q,qinv, rem64);
	}
	else if(p > 3)
	{
		// We always start with p = 2 and M-square that to get p = 3:
		MONT_SQR64(rem64,q,qinv,itmp64);
		bmap = 0;		// Init bitstring
		for(j = 0; p > 5; j++)		// j counts #bits processed
		{
			BIT_SETC(bmap,j,IS_EVEN(p));
			p = (p >> 1) + 1;
		}
		// Now do the needed powering. We always start with p = 2 and M-square that to get p = 3:
		MONT_SQR64(rem64,q,qinv,itmp64);	// rem64 has p = 2, itmp64 has p = 3
		if(p == 4) {
			MONT_MUL64(itmp64,rem64,q,qinv,rem64);
		} else if(p == 5) {
			MONT_SQR64(itmp64,q,qinv,rem64);
		} else {
			ASSERT(HERE, 0,"Bad starting value for power p!");
		}
		for(i = j-1; i >= 0; i--)
		{
			if(BIT_TEST(bmap,i)) {
				itmp64 = rem64;
				MONT_UNITY_MUL64(itmp64,q,qinv,itmp64);	// Reduce power of B by 1 in one of the 2 multiplicands...
				MONT_MUL64(itmp64,rem64,q,qinv,rem64);	// ...and multiply `em.
			} else {
				MONT_SQR64(rem64,q,qinv,rem64);
			}
		}
	}
	return rem64;
}

// Conventional positive-power version of twopmodq, returns true mod:
uint64 twopmmodq64(uint64 p, uint64 q)
{
	 int32 j,nshift;
	uint64 qhalf, qinv, x, rsqr;
	static uint64 psave = 0, zstart;
	static uint32 start_index, first_entry = TRUE;

	qhalf  = q>>1;	/* = (q-1)/2, since q odd. */

	if(first_entry || p != psave)
	{
		first_entry = FALSE;
		psave  = p;
		start_index = 64-leadz64(p)-6;
		zstart = (uint64)1 << ibits64(p,start_index,6);
	}

	nshift = trailz64(q);
	if(nshift)
	{
		ASSERT(HERE, p >= nshift, "twopmodq64 : Must add code to explicitly save offshifted low bits of modulus!");
		q >>= nshift;
		p -= nshift;	// Must also right-shift dividend by (nshift) bits; for 2^p this means subtracting nshift from p
	}
	/* q must be odd for Montgomery-style modmul to work: */
	ASSERT(HERE, (q & (uint64)1) == 1, "twopmodq64 : q must be odd!");
	qinv = (q+q+q) ^ (uint64)2;
	for(j = 0; j < 4; j++)
	{
		qinv = qinv*((uint64)2 - q*qinv);
	}

	/*...Initialize the binary powering...*/
	x = zstart;

	rsqr = radix_power64(q,qinv,2);	// Compute R^2 (mod q) in prep. for Mont-mul with initial seed...

	MONT_MUL64(x,rsqr, q,qinv, x);	// x*R (mod q) = MONT_MUL(x,R^2 (mod q),q,qinv)
	for(j = start_index-1; j >= 0; j--)
	{
		MONT_SQR64(x,q,qinv,x);
		if((p >> j) & (uint64)1)
		{
			if(x > qhalf)	/* Combines overflow-on-add and need-to-subtract-q-from-sum checks */
			{
				x = x + x;
				x -= q;
			}
			else
				x = x + x;
		}
	}
	MONT_UNITY_MUL64(x,q,qinv,x);	// Un-scale the loop output

	// If we applied an initial right-justify shift to the modulus, restore the shift to the
	// current (partial) remainder and re-add the off-shifted part of the true remainder.
	if(nshift)
	{
		x = (x << nshift);// + rem_save;
	}

	return x;
}

// This variant returns the 4 true mods, overwriting the inputs
void twopmmodq64_q4(uint64 p, uint64 *i0, uint64 *i1, uint64 *i2, uint64 *i3, uint64 *qi0, uint64 *qi1, uint64 *qi2, uint64 *qi3)
{
	int32 j;
	uint64 q0 = *i0, q1 = *i1, q2 = *i2, q3 = *i3
	, qinv0, qinv1, qinv2, qinv3
	, x0, x1, x2, x3
	, y0, y1, y2, y3;
	static uint64 psave = 0, zstart;
	static uint32 start_index, first_entry = TRUE;
	
	if(first_entry || p != psave)
	{
		first_entry = FALSE;
		psave  = p;
		start_index = 64-leadz64(p)-6;
		zstart = (uint64)1 << ibits64(p,start_index,6);
	}

	/* q must be odd for Montgomery-style modmul to work: */
	ASSERT(HERE, q1 > 1 && q1 > 1 && q2 > 1 && q3 > 1 , "modulus must be > 1!");
	ASSERT(HERE, q0 & 1 && q1 & 1 && q2 & 1 && q3 & 1 , "even modulus!");

	qinv0 = (q0+q0+q0) ^ (uint64)2;
	qinv1 = (q1+q1+q1) ^ (uint64)2;
	qinv2 = (q2+q2+q2) ^ (uint64)2;
	qinv3 = (q3+q3+q3) ^ (uint64)2;
	for(j = 0; j < 4; j++)
	{
		qinv0 = qinv0*((uint64)2 - q0*qinv0);
		qinv1 = qinv1*((uint64)2 - q1*qinv1);
		qinv2 = qinv2*((uint64)2 - q2*qinv2);
		qinv3 = qinv3*((uint64)2 - q3*qinv3);
	}
	x0 = x1 = x2 = x3 = zstart;
	y0 = radix_power64(q0,qinv0,2);	// Compute R^2 (mod q) in prep. for Mont-mul with initial seed...
	y1 = radix_power64(q1,qinv1,2);
	y2 = radix_power64(q2,qinv2,2);
	y3 = radix_power64(q3,qinv3,2);
	MONT_MUL64(x0,y0, q0,qinv0, x0);	// x*R (mod q) = MONT_MUL(x,R^2 (mod q),q,qinv)
	MONT_MUL64(x1,y1, q1,qinv1, x1);
	MONT_MUL64(x2,y2, q2,qinv2, x2);
	MONT_MUL64(x3,y3, q3,qinv3, x3);

  #ifndef YES_ASM

	for(j = start_index-1; j >= 0; j--)
	{
		MONT_SQR64_q4(x0,x1,x2,x3,q0,q1,q2,q3,qinv0,qinv1,qinv2,qinv3,y0,y1,y2,y3);
		if((p >> j) & (uint64)1)
		{
		#ifdef NOBRANCH
			x0 = y0 + y0;	x0 -= q0 & -(x0 >= q0 || x0 < y0);
			x1 = y1 + y1;	x1 -= q1 & -(x1 >= q1 || x1 < y1);
			x2 = y2 + y2;	x2 -= q2 & -(x2 >= q2 || x2 < y2);
			x3 = y3 + y3;	x3 -= q3 & -(x3 >= q3 || x3 < y3);
		#else
			x0 = y0 + y0;	if(x0 >= q0 || x0 < y0) x0 -= q0;
			x1 = y1 + y1;	if(x1 >= q1 || x1 < y1) x1 -= q1;
			x2 = y2 + y2;	if(x2 >= q2 || x2 < y2) x2 -= q2;
			x3 = y3 + y3;	if(x3 >= q3 || x3 < y3) x3 -= q3;
		#endif
		} else {
			x0 = y0;
			x1 = y1;
			x2 = y2;
			x3 = y3;
		}
	}

  #else

	j = start_index-1;
//printf("twopmodq64_q4 : x1 = %s\n", &str0[convert_uint64_base10_char(str0, x1)] );
//for(j = start_index-1; j >= 0; j--) {
	__asm__ volatile (\
	"/* Load the q0|1|2 into rbx|rsi|rdi, keeping rcx free for loop index and rax:rdx for double-width MULs: */	\n\t"\
		"movq	%[__q0],%%rbx	\n\t"\
		"movq	%[__q1],%%rsi	\n\t"\
		"movq	%[__q2],%%rdi	\n\t"\
		"/* Must load q3 as-needed due to lack of user register to store it */\n\t"\
	"/* Load the x's into r12-15: */	\n\t"\
		"movq	%[__x0],%%r12	\n\t"\
		"movq	%[__x1],%%r13	\n\t"\
		"movq	%[__x2],%%r14	\n\t"\
		"movq	%[__x3],%%r15	\n\t"\
	"/* Pure-ASM loop control: for(j = start_index-2; j >= 0; j--) */\n\t"\
		"movslq	%[__start_index], %%rcx		\n\t"\
		"subq $1,%%rcx						\n\t"\
		"test %%rcx, %%rcx					\n\t"\
		"jl LoopEnd4a		/* Skip if n < 0 */	\n\t"\
	"LoopBeg4a:								\n\t"\
	"/* SQR_LOHI_q4(x*, lo*, hi*): */	\n\t"\
		"movq	%%r12,%%rax	/* x0-3 in r8-11. */\n\t"\
		"mulq	%%rax		\n\t"\
		"movq	%%rax,%%r8 	/* lo0 */	\n\t"\
		"movq	%%rdx,%%r12	/* hi0 */	\n\t"\
		"movq	%%r13,%%rax	\n\t"\
		"mulq	%%rax		\n\t"\
		"movq	%%rax,%%r9 	/* lo1 */	\n\t"\
		"movq	%%rdx,%%r13	/* hi1 */	\n\t"\
		"movq	%%r14,%%rax	\n\t"\
		"mulq	%%rax		\n\t"\
		"movq	%%rax,%%r10	/* lo2 */	\n\t"\
		"movq	%%rdx,%%r14	/* hi2 */	\n\t"\
		"movq	%%r15,%%rax	\n\t"\
		"mulq	%%rax		\n\t"\
		"movq	%%rax,%%r11	/* lo3 */	\n\t"\
		"movq	%%rdx,%%r15	/* hi3 */	\n\t"\
		"\n\t"\
		"\n\t"\
	"/* MULL_q4((qinv*, lo*, lo*): */	\n\t"\
		"\n\t"\
		"imulq	%[__qinv0],%%r8 	\n\t"\
		"imulq	%[__qinv1],%%r9 	\n\t"\
		"imulq	%[__qinv2],%%r10	\n\t"\
		"imulq	%[__qinv3],%%r11	\n\t"\
		"\n\t"\
	"/* UMULH_q4((q*, lo*, lo*): lo0-3 in r8-11: */	\n\t"\
		"\n\t"\
		"movq	%%r8 ,%%rax	/* lo0 */\n\t"\
		"mulq	%%rbx	/* q0 in rbx; q0*lo0 in rax:rdx */\n\t"\
		"movq	%%rdx,%%r8 	/* Discard low 64 bits [rax] */\n\t"\
		"\n\t"\
		"movq	%%r9 ,%%rax	/* lo1 */\n\t"\
		"mulq	%%rsi	/* q1 in rsi; q1*lo1 in rax:rdx */\n\t"\
		"movq	%%rdx,%%r9 	/* Discard low 64 bits [rax] */\n\t"\
		"\n\t"\
		"movq	%%r10,%%rax	/* lo2 */\n\t"\
		"mulq	%%rdi	/* q2 in rdi; q2*lo2 in rax:rdx */\n\t"\
		"movq	%%rdx,%%r10	/* Discard low 64 bits [rax] */\n\t"\
		"\n\t"\
		"movq	%%r11,%%rax	/* lo3 */\n\t"\
		"mulq	%[__q3]		\n\t"\
		"movq	%%rdx,%%r11	/* Discard low 64 bits [rax] */\n\t"\
		"\n\t"\
	"/* If h < l, then calculate h-l+q < q; otherwise calculate h-l. */\n\t"\
		"subq	%%r8 ,%%r12			\n\t" /* r12 = (h-l) */\
		"leaq (%%r12,%%rbx),%%r8 	\n\t" /* r8  = (h-l)+q */\
		"cmovcq %%r8 ,%%r12	\n\t" /* if carry = true (i.e. h > l), copy source (r8 = h-l+q) into dest (r12), else leave dest = h-l. */\
		"\n\t"\
		"subq	%%r9 ,%%r13			\n\t"\
		"leaq (%%r13,%%rsi),%%r9 	\n\t"\
		"cmovcq %%r9 ,%%r13	\n\t"\
		"\n\t"\
		"subq	%%r10,%%r14			\n\t"\
		"leaq (%%r14,%%rdi),%%r10	\n\t"\
		"cmovcq %%r10,%%r14	\n\t"\
		"\n\t"\
		"movq	%[__q3],%%rdx	/* Re-use q3 several times below, so store in free register */\n\t"\
		"subq	%%r11,%%r15			\n\t"\
		"leaq (%%r15,%%rdx),%%r11	\n\t"\
		"cmovcq %%r11,%%r15	\n\t"\
		"\n\t"\
"/* If current bit of p == 1, double each output modulo q: */	\n\t"\
		"/* if((p >> j) & (uint64)1) { */	\n\t"\
		"movl	%[__p],%%eax	/* Need to follow this with load-j-into-ecx if use HLL loop control in debug mode */\n\t"\
		"shrq	%%cl,%%rax				\n\t"\
		"andq	$0x1,%%rax				\n\t"\
	"je	twopmodq64_q4_pjmp			\n\t"\
		"\n\t"\
		"movq	%%r12,%%r8 	/* r8  <- Copy of x */\n\t"\
		"movq	%%r12,%%rax	/* rax <- Copy of x */\n\t"\
		"leaq (%%r12,%%r12),%%r12	/* x+x */\n\t"\
		"subq	%%rbx,%%r8	/* r8  <- x-q */\n\t"\
		"addq	%%rax,%%r8 	/* r8 <- 2x-q */\n\t"\
		"cmovcq %%r8 ,%%r12	\n\t"/* if carry (i.e. x+x needed modding), copy source (r8 = x+x-q) into dest (r12), else leave dest = x+x. */\
		"\n\t"\
		"movq	%%r13,%%r9 	\n\t"\
		"movq	%%r13,%%rax	\n\t"\
		"leaq (%%r13,%%r13),%%r13	\n\t"\
		"subq	%%rsi,%%r9	\n\t"\
		"addq	%%rax,%%r9 	\n\t"\
		"cmovcq %%r9 ,%%r13	\n\t"\
		"\n\t"\
		"movq	%%r14,%%r10	\n\t"\
		"movq	%%r14,%%rax	\n\t"\
		"leaq (%%r14,%%r14),%%r14	\n\t"\
		"subq	%%rdi,%%r10	\n\t"\
		"addq	%%rax,%%r10	\n\t"\
		"cmovcq %%r10,%%r14	\n\t"\
		"\n\t"\
		"movq	%%r15,%%r11	\n\t"\
		"movq	%%r15,%%rax	\n\t"\
		"leaq (%%r15,%%r15),%%r15	\n\t"\
		"subq	%[__q3],%%r11	/* Weird...using rdx in place of __q3 here made timings 1-2 percent  worse. */\n\t"\
		"addq	%%rax,%%r11	\n\t"\
		"cmovcq %%r11,%%r15	\n\t"\
		"\n\t"\
	"twopmodq64_q4_pjmp:					\n\t"\
		"/* } endif((p >> j) & (uint64)1) */						\n\t"\
		"subq	$1,%%rcx	/* j-- */		\n\t"\
		"cmpq	$0,%%rcx	/* compare j vs 0 */\n\t"\
		"jge	LoopBeg4a	/* if (j >= 0), Loop */	\n\t"\
	"LoopEnd4a:							\n\t"\
		"movq	%%r12,%[__x0]	\n\t"\
		"movq	%%r13,%[__x1]	\n\t"\
		"movq	%%r14,%[__x2]	\n\t"\
		"movq	%%r15,%[__x3]	\n\t"\
		:	/* outputs: none */\
		: [__q0] "m" (q0)	/* All inputs from memory addresses here */\
		 ,[__q1] "m" (q1)	\
		 ,[__q2] "m" (q2)	\
		 ,[__q3] "m" (q3)	\
		 ,[__qinv0] "m" (qinv0)	\
		 ,[__qinv1] "m" (qinv1)	\
		 ,[__qinv2] "m" (qinv2)	\
		 ,[__qinv3] "m" (qinv3)	\
		 ,[__x0] "m" (x0)	\
		 ,[__x1] "m" (x1)	\
		 ,[__x2] "m" (x2)	\
		 ,[__x3] "m" (x3)	\
		 ,[__p] "m" (p)	\
		 ,[__j] "m" (j)	/* Only need this if debug and explicit loop enabled, but try with/sans for each version of the asm, pivk faster one. */\
		 ,[__start_index] "m" (start_index)	\
		: "cc","memory","rax","rbx","rcx","rdx","rsi","rdi","r8","r9","r10","r11","r12","r13","r14","r15"		/* Clobbered registers */\
		);

  #endif	// YES_ASM

	MONT_UNITY_MUL64(x0,q0,qinv0,*i0);	// Un-scale the loop output
	MONT_UNITY_MUL64(x1,q1,qinv1,*i1);
	MONT_UNITY_MUL64(x2,q2,qinv2,*i2);
	MONT_UNITY_MUL64(x3,q3,qinv3,*i3);

	// User needs mod-inverses returned:
	if(qi0) {
		*qi0 = qinv0;
		*qi1 = qinv1;
		*qi2 = qinv2;
		*qi3 = qinv3;
	}
	return;
}


// Specialized code to sieve an MMp factor candidate q = 2.k.Mp + 1 to as deep as 2^64:
//uint64 factor_qmmp_sieve64(uint32 p, uint64 k, uint64 imin, uint64 imax)
int main(int argc, char *argv[])
{
	uint32 p=0, pmod60, kmod60, iarg,pass=0;
	uint64 i,j,hi, k=0,imin=0,imax=0 ,nsieve,ntry, ihi,curr_p,mask,itmp64, i0,i1,i2,i3, qinv0,qinv1,qinv2,qinv3;
	uint64 tryp[16] = {0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull,0ull};
	char *cptr;
	/* Small-primes-sieve stuff:
	LEN is the number of 64-bit words in our sieving bitmap. Successive
	bits of the sieve bitmap correpond to successive odd integers, with bit 0
	corresponding to 1.
	*/
	uint32 w64,bit,idx,l,m,ncopies,regs_todo;
	const uint32 nclear=6, len=3*5*7*11*13*17, p_last_small = 17;	/* Make sure LEN = product of first NCLEAR odd primes,
																	p_last_small = largest odd prime appearing in the product. */
	const uint32 slen = len*64;	// Number of bits of sieve
	uint32 prime[] = {3,5,7,11,13,17};	/* Also need the first [nclear] odd primes - since 'const' is treated as a readonly flag
						on *variables*, use [] instead of [nclear] to avoid variable-length array init errors. */
	uint64 *temp_late = 0x0;		/* Even though we know how large to make this, it's only needed
									for data inits, so we calloc it at runtime and free it later. */
	uint64 *bit_map = 0x0;
	uint32 nprime = NUM_SIEVING_PRIME;	/* Don't make const, in case we need to reduce it so MAX_SIEVING_PRIME < (q_min = 2*p+1) */
	uint32 max_diff, isprime,sqrt_p;
	unsigned char *pdiff;	/* Compact table storing the (difference/2) between adjacent odd primes.
							http://mathworld.wolfram.com/PrimeGaps.html shows the first >256-gap at ~387 million
							and the first >512-gap at ~300 billion, so using the half-of-even-gap trick makes
							the difference between an algo which is safe for all 32-bit primes and one with a
							limit of < 2^30. */
	uint32 *offset, *startval;
	uint32 MAX_SIEVING_PRIME = 0;

	/* Allocate the arrays and initialize the array of sieving primes	*/
	temp_late = (uint64 *)calloc(len, sizeof(uint64));
	if (temp_late == NULL) {
		fprintf(stderr,"Memory allocation failure for TEMP_LATE array");
		ASSERT(HERE, 0,"0");
	}

	bit_map   = (uint64 *)calloc(len, sizeof(uint64));
	if (bit_map == NULL) {
		fprintf(stderr,"Memory allocation failure for BITMAP array");
		ASSERT(HERE, 0,"0");
	}

	pdiff = (unsigned char *)calloc(NUM_SIEVING_PRIME, sizeof(unsigned char));
	if (pdiff == NULL) {
		fprintf(stderr,"Memory allocation failure for pdiff array");
		ASSERT(HERE, 0,"0");
	}

	offset = (uint32 *)calloc(NUM_SIEVING_PRIME, sizeof(uint32));
	if (offset == NULL) {
		fprintf(stderr,"Memory allocation failure for OFFSET array");
		ASSERT(HERE, 0,"0");
	}

	startval = (uint32 *)calloc(NUM_SIEVING_PRIME, sizeof(uint32));
	if (startval == NULL) {
		fprintf(stderr,"Memory allocation failure for STARTVAL array");
		ASSERT(HERE, 0,"0");
	}

	/* pdiff[0] = 0 represents the diff between 2 and 3, so start loop at i=1: */
	ihi = curr_p = 3;
	max_diff = 0;
	for(i = 1; i < nprime; ++i)
	{
		curr_p += 2;
		++pdiff[i];
		/* Make sure (curr_p + 29) < 2^32: */
		if(curr_p > 0xffffffe3) {
			fprintf(stderr,"curr_p overflows 32 bits!");
			nprime = i;
			break;
		}
		// It's a composite - skip to next odd integer:
		if(twopmodq32(curr_p-1, curr_p) == 0) {
			--i;
			continue;
		}
		// Samllest Fermat base-2 pseudoprime = 341, so below that we can be sure it's a prime:
		if(curr_p < 341) {
			offset[i] = (curr_p >> 1);	// Initial bit-offset w.r.to sieve
			if(pdiff[i] > max_diff)
				max_diff = pdiff[i];
			continue;
		}

		/* To speed the small-prime sieve init, test divisibility by smaller primes via 64-bit gcd rather than serial modulo.
			Use that a uint64 can store the product of odd primes 3 through 53 = 16294579238595022365.
		*/
		itmp64 = 16294579238595022365ull;
		j = gcd64((uint64)curr_p, itmp64);
		if(j > 1) {	// curr_p is base-2 Fermat pseudoprime
			--i;
			continue;
		}
		l = 59;
		idx = 15;

		/* Trial-divide the next-prime candidate by all odds <= sqrt(curr_p).
		If none divide it, curr_p is prime - increment outer-loop parameters.
		If curr_p not prime, move on to next odd number; repeat until we find next odd prime.
		*/
		/* Add a small fudge factor to the result of the sqrt to make sure RO errors don't
		cause the sqrt to be one smaller than it should be once we truncate to integer.
		*/
		sqrt_p = (uint32)(sqrt(1.0*curr_p) + 0.5);
		isprime = TRUE;

		while(l <= sqrt_p)
		{
			/* It's composite (i.e. pseudoprime) - go to next-larger odd number */
			if(curr_p%l == 0) {
				isprime = FALSE;
				break;
			}
			l += (pdiff[++idx] << 1);
		}
		if(!isprime) {
			--i;
		} else {
			ihi = curr_p;
			offset[i] = (curr_p >> 1);	// Initial bit-offset w.r.to sieve
			if(pdiff[i] > max_diff) {
				max_diff = pdiff[i];
			}
		}
	}
	MAX_SIEVING_PRIME = ihi;

	printf("Using first %u odd primes; max gap = %u\n",nprime,2*max_diff);
	printf("Max sieving prime = %u\n",MAX_SIEVING_PRIME);
	ASSERT(HERE, (MAX_SIEVING_PRIME < (slen+slen)), "Max sieving prime larger than allowed!");

	/* quick way to set all the bits = 1	*/
	for(i = 0; i < len; i++)
	{
		temp_late[i] = ~(uint64)0;
	}
	/* Find multiples of small primes.	*/
	regs_todo=1;
	for(m = 0; m < nclear; m++)
	{
		curr_p = prime[m];
		regs_todo=regs_todo*curr_p;
		for(j = 0; j < regs_todo; j++)	/* number of registers to run through while zeroing multiple of (m)th prime...	*/
		{
			for(i = 0; i < 64; i++)			/* ...and run through the 64 bits of each register. */
			{
				if((curr_p>>1) == i) {
				//	printf("First multiple of %llu in bit %llu of register %llu\n", curr_p, i, j);
					for(l = (uint32)(j<<6)+i; l < (regs_todo<<6); l += curr_p)
					{
						w64=(l>>6);	/* l/64	*/
						bit=l&63;	/* l%64	*/
						temp_late[w64] &= ~((uint64)1 << bit);
					}
					goto KLOOP;
				}
			}
		}
		/* Should never reach this regular-loop-exit point: */
		fprintf(stderr,"ERROR: failed to find a multiple of prime %llu\n", curr_p);
		ASSERT(HERE, 0,"0");

	KLOOP:
		/*...Now propagate copies of length (regs_todo) bit-cleared portion of sieve
		to remaining parts of sieve.
		*/
		if(m < (nclear-1))
		{
			ncopies = prime[m+1];
			l = regs_todo;
			for(i = 2; i <= ncopies; i++, l += regs_todo)
			{
				for(j = 0; j < regs_todo; j++)
				{
					temp_late[l+j] = temp_late[j];
				}
			}
		}
	}	/* endfor(m = 0; m < nclear; m++) */
/*
	uint32 on_bits = 0;
	for(m = 0; m < len; m++)
	{
		for(j = 0; j < 64; j++)
		{
			on_bits += (temp_late[m]>>j) & 1;
		}
	}
	printf("%u ones bits of %u in master sieve.\n", on_bits, len<<6);
*/
	// Minimum argument list: [executable name] -p [value] -k [value] -imax [value]
	if(argc < 7) {
		goto HELP;
	}
	for(iarg = 1; iarg < argc; iarg += 2)
	{
		if(!strcmp((char*)"-p", argv[iarg]))
		{
			p = (uint32)strtol(argv[iarg+1],&cptr,10);
			if(*cptr || errno || p != strtol(argv[iarg+1],&cptr,10) )
			{
				printf("ERROR: can't parse value for option \"-p\"\n");
				return EXIT_FAILURE;
			}
		}
		else if(!strcmp((char*)"-k", argv[iarg]))
		{
			k = (uint64)strtod(argv[iarg+1],&cptr);
			if(*cptr || errno)
			{
				printf("ERROR: can't parse value for option \"-k\"\n");
				return EXIT_FAILURE;
			}
		}
		else if(!strcmp((char*)"-imin", argv[iarg]))
		{
			imin = (uint64)strtod(argv[iarg+1],&cptr);
			if(*cptr || errno)
			{
				printf("ERROR: can't parse value for option \"-imin\"\n");
				return EXIT_FAILURE;
			}
		}
		else if(!strcmp((char*)"-imax", argv[iarg]))
		{
			imax = (uint64)strtod(argv[iarg+1],&cptr);
			if(*cptr || errno)
			{
				printf("ERROR: can't parse value for option \"-imax\"\n");
				return EXIT_FAILURE;
			}
		}
		else
		{
			goto HELP;
		}
	}
	ASSERT(HERE, p >= 31,"p must be >= 31!");
	ASSERT(HERE, k != 0,"k must be > 0!");
	ASSERT(HERE, imax > imin,"imax must be > imin!");
	pmod60 = twopmmodq64(p, 60ull) - 1;		// For double-Mersenne factoring need M(p) mod 60
	/*
	The above routine computes 2^p (mod 60) via Montgomery-mul-based powering. That requires an odd modulus,
	so the actual powering step computes 2^(p-2) (mod 15), multiplies the result by 4 to get 2^p (mod 60),
	and subtracts 1 to get M(p) (mod 60).
	
	Note that there is a shortcut to obtaining 2^(p-2) (mod 15), namely summing the hex digits of 2^(p-2)
	(mod 15). Since the hexadecimal base 16 == 1 (mod 15), this sum gives the desired result (mod 15).
	Since 2^(p-2) i binary is just a 1 followed by (p-2) binary zeros, in base-16 it is just a leading hex
	digit d = 2^((p-2)%4) followed by a string of (p-2)/4 hexadecimal zeros, the latter of which contribute
	0 to the hex-digit sum (mod 15). Thus
	
		2^(p-2) == 2^((p-2)%4) (mod 15), whence
	
		M(p) == 4*2^((p-2)%4) - 1 (mod 60).

	For example, for p = 521 we have 2^(p-2) == 2^3 == 8 (mod 15), whence M(p) == 31 (mod 60).
	
	For p = 607 we 2^(p-2) == 2^1 == 2 (mod 15), whence M(p) == 7 (mod 60).
	
	In fact for any odd-exponent M(p) there are only the two possibilities p == 1 or 3 (mod 4), for
	which 2^(p-2) == 2^3 or 2^1 (mod 15) and M(p) == 31 or 7 (mod 60), respectively.
	
	For these two M(p) (mod 60) values the respective sets of 16 eligible k (mod 60) values of possible factors
	q = 2.k.M(p) + 1 are

		p == 1 (mod 4), M(p) == 31 (mod 60) : k (mod 60) = (any of) 0, 5, 8, 9,20,21,24,29,33,36,41,44,45,48,53,56
	
		p == 3 (mod 4), M(p) ==  7 (mod 60) : k (mod 60) = (any of) 0, 5, 8, 9,12,17,20,24,29,32,33,44,45,48,53,57 .

	M(p) == 31 or 7 (mod 60) also implies that the following check is not needed, but include it for formal completeness,
	and in case someone else modifies this code for a purpose where the above call might in fact return 0 (mod 60):
	*/
	if(pmod60 < 0) pmod60 += 60;
	kmod60 = k%60;
//	printf("pmod60, kmod60 = %u, %u\n",pmod60,kmod60);
	ASSERT(HERE, CHECK_PKMOD60(pmod60, kmod60), "K must be one of the admissible values (mod 60) for this p (mod 60)!");
	if(!imin) imin = 3;

	imin |= 0x0000000000000001ull;	// imin must be odd
	fprintf(stderr,"Searching for small divisors of MM(%u): Q( k = %llu ) in interval [%llu, %llu]...\n",p,k,imin,imax);

	/*
	For primes greater than prime[nclear], i.e. ones not precleared in the saved sieving template, we need to store
	an offset-relative-to-current-sieve-range, which gets updated every time we finish a pass through the sieve. Fo example, if
	our sieve has len = 3*5*7*11*13*17 64-bit words or len*64 = 255255*64 bits, the 0 bit of the initial pass through this corr.
	to the odd integer 1, and each complete pass through the sieve increments the odd-integer value corr. to each sieve
	bit by 2*len*64 = 2*510510*64 = 32672640. For an odd prime p the initial bit-offset w.r.to bit 0 is just p/2 (truncating divide).
	On each pass through the sieve this offset gets incremented by [smallest multiple of p > 32672640] - 32672640
	
	For example, the small prime p = 101 has initial offset = 101/2 = 50, i.e. the lowest sieve bit it zeros is bit 50 (of <0:63>)
	of the lowest 64-bit sieve word (= bit 50 of word 0). It also zeros bit 50+101 = 151 (bit 23 of word 2), bit 50+2*101 (bit 60
	of word 3), etc. The smallest multiple of p=101 which is > 2*len*64 is p*ceiling(2*len*64/p) = 32672692 = (32672640 + 52). Thus
	on this first pass through the sieve the bit-offset for p=101 increased by 2 (from 50 to 52). We don't need yet another table
	to store these offset-increments-per-pass; rather we just compute the initial offsets and uodate them per sieve pass as follows:
	
		slen := len*64;	// Length of sieve in terms of odd numbers encoded by its bits, from one pass to the next
		if(p < 2*slen) {
			starting with current offset for this p, zero every (p)th bit of current copy of small-primes-precleared sieve.
			As soon as the resulting bit index > slen, stop and update the next-pass-offset via offset = (bit index) - slen.
		} else {
			if((current offset for this p) < slen) {
				clear the corr. sieve bit and update offset += p - slen.
			} else {
				update offset -= slen.
			}
		}
	
	We have the added complication of an arbitrarily large odd-valued starting integer (imin) here.	*** LATER ***
	*/

	// Init bit_map = Copy of small-primes-precleared sieve (size = 8*len bytes);
	memcpy(bit_map, temp_late, (len << 3));

	// If starting Pass > 0, Compute offsets for "jump-in" sieving, (p/2) - pass*slen (mod p):
	pass = (imin>>1)/slen;

	if(pass > 0) {
	//	printf("Starting with Pass %u\n",pass);
		/*
		On each pass thru sieve the per-prime sieve-bit offset gets incremented by ( p - (slen % p) ) (mod p).
		We have slen = len*64 = 64*3*5*7*11*13*17 = 16336320;
		For example:
		p = 17333; slen % p =  8634; p - (slen % p) =  8699
		p = 17341; slen % p =  1098; p - (slen % p) = 16243
		p = 17351; slen % p =  9029; p - (slen % p) =  8322
		p = 17359; slen % p =  1501; p - (slen % p) = 15858
		p = 17377; slen % p =  1940; p - (slen % p) = 15437
		p = 17383; slen % p = 13683; p - (slen % p) =  3700
		p = 17387; slen % p =  9927; p - (slen % p) =  7460
		p = 17389; slen % p =  8049; p - (slen % p) =  9340
		p = 17393; slen % p =  4293; p - (slen % p) = 13100
		
		And from Pass 0 to Pass 1 we see the increments in the right column get added to the current offsets, modulo the prime in question:
		Pass = 1; Init sieve...
		p = 17333; Updating offset from 32 to 8731
		p = 17341; Updating offset from 7572 to 6474
		p = 17351; Updating offset from 16997 to 7968
		p = 17359; Updating offset from 7178 to 5677
		p = 17377; Updating offset from 6748 to 4808
		p = 17383; Updating offset from 12391 to 16091
		p = 17387; Updating offset from 16153 to 6226
		p = 17389; Updating offset from 645 to 9985
		p = 17393; Updating offset from 4403 to 110
		
		The initial (pass 0) offset for each p is given simply by (p/2).
		Thus, if the user specifies a lower bound b0 for the sieving, in order to "jump into the sieving" there,
		we need to do as follows:
		
		1. Compute the initial (pass 0) offsets for each small sieving prime, (p/2);
		
		2. Compute how many completed sieve passes b0 corresponds to, via pass = b0/(2*slen);
		
		3. For each p, compute the offset which would result from (pass) complete sieving passes, via (p/2) - pass*slen (mod p).
		*/
		curr_p = p_last_small;
		for(m = nclear; m < nprime; m++)
		{
			curr_p += (pdiff[m] << 1);
			offset[m] = ((curr_p>>1) + (uint64)pass*(curr_p - slen%curr_p) )%curr_p;
		}
	}
	// Note that curr_p for this sieve-bit-clearing will never exceed 32 bits.
	curr_p = p_last_small;
	for(m = nclear; m < nprime; m++)
	{
		curr_p += (pdiff[m] << 1);
		l = offset[m];
		DBG_ASSERT(HERE, (bit_map[l>>6]>>(l&63)) & 0x1, "First occurrence of each prime must hit uncleared bit!");
	//	if(curr_p < (slen+slen)) {
			// starting with 2nd offset for this p (1st one is p itself), zero every (p)th bit of current copy of small-primes-precleared sieve.
			l += curr_p;	// Skip p itself
			while(l < slen) {
				w64=(l>>6);	/* Current 64-bit sieve word */
				bit=l&63;	/* l%64	*/
				mask = ~((uint64)1 << bit);
				bit_map[w64] &= mask;
				l += curr_p;
			}
			// As soon as the resulting bit index > slen, stop and update the next-pass-offset via offset = (bit index) - slen.
			offset[m] = l - slen;
	/*	} else {
			l = offset[m];
			if(l < slen) {
				// clear the corr. sieve bit and update offset += p - slen:
				w64=(l>>6);	// Current 64-bit sieve word
				bit=l&63;	// l%64
				mask = ~((uint64)1 << bit);
				bit_map[w64] &= mask;
				l += curr_p;
				offset[m] = l - slen;
			} else {
				offset[m] -= slen;
			}
		}	*/
	}

	// Find starting location in small-primes sieve:
	l = (imin>>1) % slen;
	w64=(l>>6) % len;	/* Current 64-bit sieve word */
	bit=l&63;	/* l%64	*/
	l = (w64<<6) + bit;
	nsieve = ntry = 0;	// Number of current odds which passed the small-p sieve and get queued up for base-2 PRP testing

	for(i = imin; /* i < imax */; i += 2, l++)
	{
		// Init for a fresh pass through the sieve:
		if(l == slen) {
			pass++;	// Keep track of #passes through the bit-cleared sieve
			// bit_map = Copy of small-primes-precleared sieve (size = 8*len bytes):
		#if 0
			memcpy(bit_map, temp_late, (len << 3));	// Oddly, direct loop-copy below faster
		#else
			for(j = 0; j < len; j++) {
				bit_map[j] = temp_late[j];
			}
		#endif
			curr_p = p_last_small;
			for(m = nclear; m < nprime; m++)
			{
				curr_p += (pdiff[m] << 1);
				l = offset[m];
				// Here we assume (curr_p < (slen+slen)):
				// starting with current offset for this p, zero every (p)th bit of current copy of small-primes-precleared sieve.
				while(l < slen) {
					w64=(l>>6);	/* Current 64-bit sieve word */
					bit=l&63;	/* l%64	*/
					mask = ~((uint64)1 << bit);
					bit_map[w64] &= mask;
					l += curr_p;
				}
				// As soon as the resulting bit index > slen, stop and update the next-pass-offset via offset = (bit index) - slen.
				offset[m] = l - slen;
			}
			l = 0;	// Reset this last, since used l as an index in above offset-updates
		}
		DBG_ASSERT(HERE, l < slen, "l out of range!");
		w64=(l>>6);	/* Current 64-bit sieve word */
		bit=l&63;	/* l%64	*/
		if(((bit_map[w64] >> bit) & 0x1) == 1) {
			tryp[ntry++] = i;
		}

		// If have enough for a full batch of 4, try 'em out:
		if(ntry == 4) {
			ntry = 0;	nsieve += 4;
			i0 = tryp[0];
			i1 = tryp[1];
			i2 = tryp[2];
			i3 = tryp[3];

			// Compute q = 2.k.Mp+1 (mod tryp[]) via first Mp (mod tryp[]) and then doing the doubling, mul-by-k and +1:

			// Returns 2^p (mod i0-3) in i0-3; also return the 4 mod-inverses for reuse in k* step below
			twopmmodq64_q4(p,&i0,&i1,&i2,&i3 ,&qinv0,&qinv1,&qinv2,&qinv3);

			// Fast way of doing 2*M(p)-1 (mod-tryp[]):
			i0 += i0 - tryp[0] - 2;
			i1 += i1 - tryp[1] - 2;
			i2 += i2 - tryp[2] - 2;
			i3 += i3 - tryp[3] - 2;
			// It is possible that i0-3 < 2 which would need two re-adds of tryp[[, but that can only happen if tryp not a factor:
			i0 += (-((int64)i0 < 0)) & tryp[0];
			i1 += (-((int64)i1 < 0)) & tryp[1];
			i2 += (-((int64)i2 < 0)) & tryp[2];
			i3 += (-((int64)i3 < 0)) & tryp[3];

			// Compute k.(2p (mod curr_p)) + 1, will do final (mod curr_p) in batch mode:
			MUL_LOHI64(k,i0, i0, hi);
			MUL_LOHI64(k,i1, i1, hi);
			MUL_LOHI64(k,i2, i2, hi);
			MUL_LOHI64(k,i3, i3, hi);
			// Spot-check to ensure product < 2^64:
			ASSERT(HERE, hi == 0, "No > 64-bit suppor yet!");
			i0 = (i0   + 1)*qinv0;	i1 = (i1   + 1)*qinv1;	i2 = (i2   + 1)*qinv2;	i3 = (i3   + 1)*qinv3;

			// Length-1-4-way specialized inlining of mi64_is_div_by_scalar64; also combine i += 1 and i *= qinv steps:
			MULH64(tryp[0],i0, i0);	MULH64(tryp[1],i1, i1);	MULH64(tryp[2],i2, i2);	MULH64(tryp[3],i3, i3);

			j = (i0 == 0) + ((i1 == 0) << 1) + ((i2 == 0) << 2) + ((i3 == 0) << 3);
			// If more than one of current quarter divides q (unlikely), we print the smallest divisor:
			if(j) {
				printf("MM(%u): Q( k = %llu ) has a small factor: %llu\n", p, k, tryp[trailz32(j)]);
				goto DONE;
			}

			// Put loop-exit check here, to make sure we test at least through imax:
			if(tryp[3] >= imax) break;
			// Print some basic data for every 2^30th candidate (roughly every billionth one):
			if((nsieve & 0x3fffffff) == 0) {
				printf("At n = %llu,", tryp[3]);
				if(twopmmodq64(tryp[3]-1, tryp[3]) == 1) {
					printf(" which is base-2 PRP\n");
				} else {
					printf(" which is composite\n");
				}
			}
		}
	}
	fprintf(stderr,"MM(%u): Q( k = %llu ) has no factors in interval [%llu, %llu].\n",p,k,imin,imax);

DONE:
	fprintf(stderr,"Tried %llu ints which passed small-p sieve..\n",nsieve);
	return EXIT_SUCCESS;

HELP:
	printf("ERROR: Bad argument list! Invoke with -p [MMp exponent] -k [factor candidate k] -imin [lower sieving bound (default = 0)] -imax [upper bound]\n");
	return EXIT_FAILURE;
}

